﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Common;
using System.Data.OracleClient;
using System.Data;
using System.Globalization;

using DbSharp.Framework.DbObject ;
using DbSharp.Utils.Db;
using DbSharp.Framework.Common;
using DbSharp.Framework.Driver.FieldPersistence;

namespace DbSharp.Framework.Driver
{
    /// <summary>
    /// Mapping of Oracle data types and Microsoft SQL Server data types 
    /// http://msdn.microsoft.com/en-us/library/ms151817.aspx
    /// </summary>
	/*
	 Currently, System.Data.OracleClient cannot support the PL/SQL collection data type 
	 of procedure's parameters, such varray or table type. But ODP driver has already 
	 supported this feature, have a look at http://www.oracle.com/technology/oramag/oracle/07-jan/o17odp.html
	*/
    public class OracleTestPlugin : ITestDataSource
    {
        #region Private method

        /// <summary>
        /// remove all semicolon symbols tailed the sql statement
        /// </summary>
        /// <param name="sql"></param>
        /// <returns></returns>
        private string RemoveTailedSemiColons(string sql)
        {
            return SqlTextHelper.RightTrimCharacter(sql, ';');
        }

        
        private int GetArgumentType(string argumentTypeText)
        {
            OracleType parameterType;
            string upperTypeStr = argumentTypeText.Trim().ToUpper();
            switch (upperTypeStr)
            {
                case "ORA_CHAR":
                    parameterType = OracleType.Char;
                    break;
                case "ORA_VARCHAR":
                case "ORA_VARCHAR2":
                    parameterType = OracleType.VarChar;
                    break;
                case "ORA_NCHAR":
                    parameterType = OracleType.NChar;
                    break;
                case "ORA_NVARCHAR":
                case "ORA_NVARCHAR2":
                    parameterType = OracleType.NVarChar;
                    break;
                case "ORA_NUMBER":
                    parameterType = OracleType.Number;
                    break;
                case "ORA_DATETIME":
                    parameterType = OracleType.DateTime;
                    break;
                case "ORA_TIMESTAMP":
                    parameterType = OracleType.Timestamp;
                    break;
                case "ORA_TIMESTAMPLOCAL":
                    parameterType = OracleType.TimestampLocal;
                    break;
                case "ORA_TIMESTAMPWITHTZ":
                    parameterType = OracleType.TimestampWithTZ;
                    break;
                case "ORA_INTERVALDAYTOSECOND":
                    parameterType = OracleType.IntervalDayToSecond;
                    break;
                case "ORA_INTERVALYEARTOMONTH":
                    parameterType = OracleType.IntervalYearToMonth;
                    break; 
                case "ORA_INT":
                    parameterType = OracleType.Int32;
                    break;
                case "ORA_FLOAT":
                case "ORA_REAL":
                    parameterType = OracleType.Float;
                    break;
                case "ORA_DOUBLE": 
                    parameterType = OracleType.Double;
                    break;
                case "ORA_ROWID":
                    parameterType = OracleType.RowId;
                    break;
                case "ORA_CURSOR":
                    parameterType = OracleType.Cursor;
                    break;
                default:
                    parameterType = OracleType.VarChar;
                    break;
            }
            return (int)parameterType;
        }

        #endregion


        #region ITestDbSource Members

        public bool HoldResultsetMustWithNamedArgInSP
        {
            get { return true; }
        }


        public bool TransactionSupported
        {
            get { return true; }
        }

        public DateTime GetDbServerTime(DbConnection conn)
        {
            string sql = "Select sysdate from dual ";
            DbCommand cmd = new OracleCommand();
            cmd.Connection = conn;
            cmd.CommandType = CommandType.Text;
            cmd.CommandText = sql;
            DateTime result = (DateTime)cmd.ExecuteScalar();
            return result;
        }
        
        public bool IsRecordsetTypeArgument(string argumentTypeText)
        {
            if (GetArgumentType(argumentTypeText) == (int)OracleType.Cursor)
                return true;
            else
                return false;
        }

        public bool IsRecordsetTypeParameter(DbParameter param)
        {
            if (param == null)
                return false;
            else if (((OracleParameter)param).OracleType == OracleType.Cursor)
                return true;
            else
                return false;
        }

        public DbConnection CreateConnection(string connString)
        {
            var cnn = new OracleConnection();
            cnn.ConnectionString = connString;
            return cnn;
        }


        public string NormalizeArgumentName(string argumentName)
        {
            return SqlTextHelper.LeftTrimCharacter(argumentName, ':');
        }


        public DbCommand PrepareCommand(IStatementDeclaration statement)
        {
            DbCommand cmd = new OracleCommand();
            cmd.CommandType = statement.CommandType;
      
            IList<StatementArgument> listArgumentObj = statement.ListArgument;
            foreach (var argumentObj in listArgumentObj)
            {
                if (argumentObj.Direction==ArgumentDirection.ClientSide)
                {
                    //Clientside parameter will not added into ParameterList for ever
                    continue;
                }
                else if (argumentObj.IsSqlStmtResultArg)
                {
                    //It means this argument is virtual argument, not the real argument
                    // So the argument will not be added into ParameterList 
                    continue;
                }
                else
                {
                    //Oracle always use named arguments to hold resultsets, so we need add these return type arguments into parameterlist
                } 

                OracleParameter parameter = new OracleParameter();
                parameter.Direction = ArgumentDirectionHelper.ConvertToParamDirection(argumentObj.Direction);
                parameter.OracleType = (OracleType)GetArgumentType(argumentObj.TypeStr);
                parameter.Size = argumentObj.Size;
                parameter.ParameterName = NormalizeArgumentName(argumentObj.Name); 

                cmd.Parameters.Add(parameter);
            }

            return cmd;
        }

        public void UpdateCommandText(DbCommand cmd, string newCmdText)
        {
            cmd.CommandText = newCmdText.Trim();
            //make sure the sql statement does not end with semicolon
            if (cmd.CommandType == CommandType.Text)
            {
                cmd.CommandText = RemoveTailedSemiColons(cmd.CommandText);
            }
        }

        public void AssignParameterValue(DbParameter param, string value, string dateTimeFormat) 
        {
            OracleParameter parameter = (OracleParameter)param;

            if ((parameter.OracleType == OracleType.DateTime) ||
                (parameter.OracleType == OracleType.Timestamp) ||
                (parameter.OracleType == OracleType.TimestampLocal) ||
                (parameter.OracleType == OracleType.TimestampWithTZ))
            {
                if (string.IsNullOrEmpty(dateTimeFormat))
                    parameter.Value = DateTimeParser.ParseUtc(value);
                else
                {
                    parameter.Value = DateTimeParser.ParseExactUtc(value, dateTimeFormat);
                }
            }
            else if (parameter.OracleType == OracleType.IntervalDayToSecond)
                parameter.Value = OracleTimeSpan.Parse(value);
            else if (parameter.OracleType == OracleType.IntervalYearToMonth)
                parameter.Value = OracleMonthSpan.Parse(value);
            else
                parameter.Value = value;
        }


  
      


        private IFieldPersistence m_FieldPersistenceService = new OracleFieldPersistence();
        public IFieldPersistence FieldPersistenceService
        {
            get { return m_FieldPersistenceService; }
        }


        #endregion

    }
}
